import java.io.*;
import java.util.*;
import java.awt.Color;

import javax.vecmath.*;

public class UniGrafixParser
{
    private HashMap definitions;
    private HashMap colors;
    private HashMap rooms;
    private HashSet triangles;
    private float minX, minY, minZ;
    private float maxX, maxY, maxZ;

    public UniGrafixParser(String filename)
    {
	// definitions = new HashMap(11); // ignore defs for now
	init();

        try { readFile(filename); }
        catch(FileNotFoundException fnfe)
            {
                System.out.println("File not found: " + filename);
                System.exit(-1);
            }
        catch(IOException ie)
            {
                System.out.println("Error loading file: " + filename);
                System.exit(-1);
            }

        expandRooms();
    }

    private void init()
    {
        definitions = new HashMap(19);
        colors = new HashMap(19);
        rooms = new HashMap(101);
        triangles = new HashSet(997); // some nice big prime number
	minX = Float.POSITIVE_INFINITY;
	minY = Float.POSITIVE_INFINITY;
	minZ = Float.POSITIVE_INFINITY;
	maxX = Float.NEGATIVE_INFINITY;
	maxY = Float.NEGATIVE_INFINITY;
	maxZ = Float.NEGATIVE_INFINITY;
	colors.put("Java3DDefault", new Color3f(1.0f,1.0f,1.0f));
    }

    // Only top level instances are those that count!
    public void readFile(String filename)
	throws FileNotFoundException, IOException
    {
	System.out.println("Loading: " + filename);

	String currentLine;
	String currentValue;
	StringTokenizer strtok;
	FileReader fileReader = new FileReader(filename);
	BufferedReader bufferedFileIn = new BufferedReader(fileReader);

	currentLine = bufferedFileIn.readLine();
	while (currentLine != null)
            {             
                strtok = new StringTokenizer(currentLine, " ;");
                currentValue = strtok.nextToken();
                
                if ( currentValue.equals("c_rgb") )
                    parseColor(strtok);
                else if ( currentValue.equals("def") )
                    parseDef(bufferedFileIn, strtok);
                else if ( currentValue.equals("i") )
                    {
                        String instanceName = strtok.nextToken();
                        instanceName = instanceName.substring
                            (1, instanceName.length() - 1); // eat the ( )
                        Room topLevelRoom =
                            (Room)(definitions.get(instanceName));
                        if (topLevelRoom != null)
                            rooms.put(instanceName,
                                      definitions.get(instanceName));
                        else
                            System.out.println
                                ("Definition does not exist: " + instanceName);
                    }
                else
                    System.out.println("Invalid UG line: " + currentLine);

                currentLine = bufferedFileIn.readLine();
            }
	
	System.out.println("Load Complete.");
    }

    private void parseColor(StringTokenizer strtok)
    {
	String name = strtok.nextToken();
	float red = Float.parseFloat(strtok.nextToken());
	float green = Float.parseFloat(strtok.nextToken());
	float blue = Float.parseFloat(strtok.nextToken());
	addColor(name, new Color3f(red,green,blue));
    }

    private void addColor(String name, Color3f color)
    {
        colors.put(name, color);
    }

    private void parseDef
        (BufferedReader bufferedFileIn, StringTokenizer strtok)
        throws IOException
    {
        String defName = strtok.nextToken();
        StringTokenizer deftok;
        HashMap vertices = new HashMap(101);
        Room currentRoom = new Room(defName);
        String currentLine = bufferedFileIn.readLine();
        String currentValue;
        boolean insideDef = true;
        while (insideDef)
            {
                deftok = new StringTokenizer(currentLine, " ;");
                currentValue = deftok.nextToken();
                
                if (currentValue.equals("v"))
                    parseVertex(deftok, vertices);
                else if (currentValue.equals("f"))
                    parseFace(deftok, vertices, currentRoom);
                else if (currentValue.equals("end"))
                {
                    definitions.put(defName, currentRoom);
                    return;
                }
                else if (currentValue.equals("i"))
                    {
                        String instanceName = deftok.nextToken();
                        instanceName = instanceName.substring
                            (1, instanceName.length() - 1);
                        Room instanceRoom =
                            (Room)(definitions.get(instanceName));
                        if (instanceRoom != null)
                            currentRoom.addRoom(instanceRoom);
                        else
                            System.out.println
                                ("Definition does not exist: " + instanceName);
                    }

                currentLine = bufferedFileIn.readLine();
            }
        
    }
    
    private void parseVertex(StringTokenizer strtok, HashMap vertices)
    {
        String name = strtok.nextToken();
        float x = Float.parseFloat(strtok.nextToken());
        float y = Float.parseFloat(strtok.nextToken());
        float z = Float.parseFloat(strtok.nextToken());
        vertices.put(name, new Point3f(x,y,z));
    }

    private void parseFace
        (StringTokenizer strtok, HashMap vertexMap, Room currentRoom)
    {
        int numVertices = 0;
        ArrayList vertices = new ArrayList(3);
        String name = strtok.nextToken();
        String openParen = strtok.nextToken(); // eat the (

        if ( !(openParen.equals("(")) )
            throw new IllegalArgumentException("Improperly formatted file");
        else
            {
                String currentToken = strtok.nextToken();
                while ( !(currentToken.equals(")")) )
                    {
                        ++numVertices;
                        vertices.add(vertexMap.get(currentToken));
                        currentToken = strtok.nextToken();
                    }
		
                String colorName;
                try
                    {
                        colorName = strtok.nextToken();
                    }
                catch (NoSuchElementException nee)
                    {
                        colorName = "Java3DDefault";
                    }
		
                // TODO: Triangulate arbitrary polygon
                if (numVertices < 3)
                    throw new IllegalArgumentException
                        ("Improperly formatted file");
                else if (numVertices == 3)
                    {
                        Triangle3f t = new Triangle3f
                            ((Point3f)(vertices.get(0)),
                             (Point3f)(vertices.get(1)),
                             (Point3f)(vertices.get(2)),
                             colorName);
                        currentRoom.addTriangle(t);
                    }
                else if (numVertices == 4)
                    {
                        Point3f v0 = (Point3f)(vertices.get(0));
                        Point3f v1 = (Point3f)(vertices.get(1));
                        Point3f v2 = (Point3f)(vertices.get(2));
                        Point3f v3 = (Point3f)(vertices.get(3));
			
                        Triangle3f t1 = new Triangle3f(v0, v2, v1, colorName);
                        Triangle3f t2 = new Triangle3f(v0, v3, v2, colorName);
                        currentRoom.addTriangle(t1);
                        currentRoom.addTriangle(t2);
                    }
            }
    }

    private void expandRooms()
    {
        LinkedList currentQueue = new LinkedList(rooms.keySet());
        LinkedList nextQueue = new LinkedList();

        String currentRoomName;
        Room currentRoom;
        HashMap currentRoomMap;
        while ( !(currentQueue.isEmpty()) )
            {
                for (Iterator itr = currentQueue.iterator(); itr.hasNext();)
                    {
                        currentRoomName = (String)(itr.next());
                        currentRoom = (Room)(rooms.get(currentRoomName));
                        currentRoomMap = currentRoom.getRooms();
                        rooms.putAll(currentRoomMap);
                        nextQueue.addAll(currentRoomMap.keySet());
                    }
                currentQueue = nextQueue;
                nextQueue = new LinkedList();
            }
    }

    public Color3f getColorByName(String colorName)
    {
	Object c = colors.get(colorName);
	if (c != null)
	    return (Color3f)(c);
	else
	    return new Color3f(1.0f,1.0f,1.0f);
    }

    public HashMap getRooms()
    {
        return rooms;
    }

    public BoundingBox3f getBoundingBox()
    {
        LinkedList boundingBoxes = new LinkedList();
        Room currentRoom;
        for (Iterator itr = rooms.values().iterator(); itr.hasNext();)
            {
                currentRoom = (Room)(itr.next());
                boundingBoxes.add(currentRoom.getBoundingBox());
            }
        return BoundingBox3f.calculateFromBoundingBoxes(boundingBoxes);
    }

    public HashSet getTriangles()
    {
        HashSet triangleSet = new HashSet();
        Room currentRoom;
        for (Iterator itr = rooms.values().iterator(); itr.hasNext();)
        {
            currentRoom = (Room)(itr.next());
            triangleSet.addAll(currentRoom.getTriangles());
        }
        return triangleSet;
    }

    public static void main(String[] args)
    {
        UniGrafixParser ugp = new UniGrafixParser(args[0]);
        System.out.println(ugp.getBoundingBox());
    }
}
