// Java Utilities
import java.util.*;

// AWT/Swing
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

// Java3D
import javax.media.j3d.*;
import javax.vecmath.*;

// Java3D Utility Classes
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.behaviors.keyboard.*;
import com.sun.j3d.utils.geometry.*;

// Jama Matrix Package
import Jama.*;

// TODO: probably subclass the beacon renderer off the
// more general renderer
// the keys for the spheres that represent listeners 1 and 2 are
// "Listener 1" and "Listener 2"
public class UniGrafixRenderer
{
    private Canvas3D currentCanvas;
    private SimpleUniverse universe;

    private BranchGroup scene;
    private ViewingPlatform viewingPlatform;
    private TransformGroup roomTransformGroup;
    private Transform3D roomTransform;

    // Behavioral Elements
    private MouseRotate mouseRotate;
    private MouseTranslate mouseTranslate;
    private MouseZoom mouseZoom;

    // Internal Representation Elements
    private float minX, minY, minZ, maxX, maxY, maxZ;
    private Vector3f roomCentroid;
    private Shape3D roomShape;
    private boolean axesState;
    private Switch axesSwitch;
    private Shape3D axes;
    private boolean gridState;
    private Switch gridSwitch;
    private Shape3D roomPlane;
    private boolean beaconsState;
    private Switch beaconsSwitch;
    private boolean listenerOneRadiiState;
    private Switch listenerOneRadiiSwitch;
    private boolean listenerTwoRadiiState;
    private Switch listenerTwoRadiiSwitch;
    private Shape3D lever;
    private Shape3D laserBeam;
    // maps names (Strings) to primitives/their transformations
    private HashMap primitives;
    private HashMap primitiveTransformGroups;
    private HashMap shapes;
    private HashMap shapeTransformGroups;
    private boolean rendered;

    // TODO: maybe put the radii in a separate collection
    
    // Constructor
    public UniGrafixRenderer()
    {
	axesState = true;
	gridState = true;
	beaconsState = true;
	listenerOneRadiiState = false;
	listenerTwoRadiiState = false;
	rendered = false;
	currentCanvas = new Cricket3DCanvas();
    }

    // ============== HashMap manipulation utility functions =============
    public void addPrimitiveToRoom(String name, Primitive p)
    {
	TransformGroup tg = new TransformGroup(new Transform3D());
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

	primitives.put(name, p);
	primitiveTransformGroups.put(name, tg);

	tg.addChild(p);
	roomTransformGroup.addChild(tg);
    }

    public void addShapeToRoom(String name, Shape3D s)
    {
	TransformGroup tg = new TransformGroup(new Transform3D());
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

	shapes.put(name, s);
	shapeTransformGroups.put(name, tg);

	tg.addChild(s);
	roomTransformGroup.addChild(tg);
    }

    public void addSwitchToRoom(Switch s)
    {
	roomTransformGroup.addChild(s);
    }

    public void addPrimitiveToRoomWithSwitch
	(String name, Primitive p, Switch s)
    {
	TransformGroup tg = new TransformGroup(new Transform3D());
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
	tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

	primitives.put(name, p);
	primitiveTransformGroups.put(name, tg);

	tg.addChild(p);
	s.addChild(tg);
    }

    public Primitive getPrimitiveByName(String name)
    {
	return (Primitive)(primitives.get(name));
    }

    public Shape3D getShapeByName(String name)
    {
	return (Shape3D)(shapes.get(name));
    }

    public TransformGroup getPrimitiveTransformGroupByName(String name)
    {
	return (TransformGroup)(primitiveTransformGroups.get(name));
    }

    public TransformGroup getShapeTransformGroupByName(String name)
    {
	return (TransformGroup)(shapeTransformGroups.get(name));
    } 

    // guarantees not to change t
    public void applyPrimitiveTransformByName(String name, Transform3D t)
    {
	TransformGroup tg
	    = (TransformGroup)(primitiveTransformGroups.get(name));
	if (tg == null)
	    throw new RuntimeException("Node not found!");
	else
	    {
		Transform3D oldTransform = new Transform3D();
		// put the old transform in, idiocy
		tg.getTransform(oldTransform);
		Transform3D newTransform = new Transform3D(t);
		newTransform.mul(oldTransform);
		tg.setTransform(newTransform);
	    }
    }

    public void setPrimitiveTransformByName(String name, Transform3D t)
    {
	TransformGroup tg
	    = (TransformGroup)(primitiveTransformGroups.get(name));
	if (tg == null)
	    throw new RuntimeException("Node not found!");
	else
	    {
		tg.setTransform(t);
	    }
    }

    public Vector3d getPrimitivePositionByName(String name)
    {
	TransformGroup tg =
	    (TransformGroup)(primitiveTransformGroups.get(name));

	if (tg == null)
	    throw new RuntimeException("Node not found!");
	else
	    {
		Transform3D t = new Transform3D();
		tg.getTransform(t);
		Vector3d position = new Vector3d();
		t.get(position);
		return position;
	    }
    }

    public void setPrimitivePositionByName(String name, Vector3f t)
    {
	TransformGroup tg =
	    (TransformGroup)(primitiveTransformGroups.get(name));

	if (tg == null)
	    throw new RuntimeException("Node not found!");
	else
	    {
		Transform3D translation = new Transform3D();
		translation.setTranslation(t);
		tg.setTransform(translation);
	    }
    }

    // ============ End HashMap manipulation utility functions ===========

    public void render(UniGrafixParser ugp, HashMap beaconLocations)		      
    {
	initScene();
	readScene(ugp);
	// transformScene();
	initAxes();
	initGrid(24);
	initListenerRadii(beaconLocations);
	initLever();
	initLaserBeam();
	drawCursor();
	drawListeners();
	finishScene();
	rendered = true;
    }

    private void initScene()
    {
	primitives = new HashMap(5);
	primitiveTransformGroups = new HashMap(5);
	shapes = new HashMap(5);
	shapeTransformGroups = new HashMap(5);
	
	/*
	addKeyListener(new KeyAdapter()
	    {
		Shape3D l;
		LineArray la;
		Point3f p;
		public void keyPressed(KeyEvent e)
		{
		    char c = e.getKeyChar();
		    Transform3D apply;
		    switch(c)
			{
			case 'r':
			    apply = new Transform3D();
			    apply.setTranslation(new Vector3f(0.0f,0.0f,2.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			case 'f':
			    apply = new Transform3D();
			    apply.setTranslation
				(new Vector3f(0.0f,0.0f,-2.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			case 'a':
			    apply = new Transform3D();
			    apply.setTranslation
				(new Vector3f(-2.0f,0.0f,0.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			case 'd':
			    apply = new Transform3D();
			    apply.setTranslation
				(new Vector3f(2.0f,0.0f,0.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			case 'w':
			    apply = new Transform3D();
			    apply.setTranslation
				(new Vector3f(0.0f,2.0f,0.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			case 's':
			    apply = new Transform3D();
			    apply.setTranslation
				(new Vector3f(0.0f,-2.0f,0.0f));
			    applyPrimitiveTransformByName("cursor", apply);
			    break;
			default:
			    break;
			}
		    TransformGroup tg;
		    Transform3D t = new Transform3D();
		    Vector3f position = new Vector3f();
		    tg = (getPrimitiveTransformGroupByName("cursor"));
		    tg.getTransform(t);
		    t.get(position);
		}
	    });
	*/
	scene = new BranchGroup();
	roomTransformGroup = new TransformGroup();
    }

    private void readScene(UniGrafixParser ugp)
    {
	Collection triangleObjects = ugp.getTriangles();
	int numTriangles = triangleObjects.size();
	TriangleArray ta = new TriangleArray( 3*numTriangles,
					      TriangleArray.COORDINATES
					      | TriangleArray.COLOR_3 );
	Triangle3f currentTriangle;
	Point3f v1, v2, v3;
	Color3f currentColor;
	int i = 0;
	for(Iterator itr=triangleObjects.iterator(); itr.hasNext();)
	    {
		currentTriangle = (Triangle3f)(itr.next());
		v1 = currentTriangle.v1;
		v2 = currentTriangle.v2;
		v3 = currentTriangle.v3;
		currentColor = ugp.getColorByName(currentTriangle.colorName);

		ta.setCoordinate(i, v1);
		ta.setColor(i,currentColor);
		ta.setCoordinate(i+1, v2);
		ta.setColor(i+1,currentColor);
		ta.setCoordinate(i+2, v3);
		ta.setColor(i+2,currentColor);
		i += 3;
	    }

        BoundingBox3f bbox = ugp.getBoundingBox();
	minX = bbox.minX;
	minY = bbox.minY;
	minZ = bbox.minZ;
	maxX = bbox.maxX;
	maxY = bbox.maxY;
	maxZ = bbox.maxZ;

	roomCentroid = new Vector3f
	    ( (minX + maxX) / 2,
	      (minY + maxY) / 2,
	      (minZ + maxZ) / 2 );

	Appearance appr = new Appearance();
	PolygonAttributes polyattr =
	    new PolygonAttributes(PolygonAttributes.POLYGON_FILL,
				  PolygonAttributes.CULL_NONE,
				  0.0f);
	appr.setPolygonAttributes(polyattr);

	roomShape = new Shape3D(ta, appr);
	scene.addChild(roomShape);
    }

    private void initAxes()
    {
	LineArray la = new LineArray(6,
				     LineArray.COORDINATES |
				     LineArray.COLOR_3);
	float maxCoord = Math.max(Math.max(maxX,maxY),maxZ);

	// x-axis
	la.setCoordinate(0, new Point3f(0, 0.0f, 0.0f));
	// la.setCoordinate(0, new Point3f(0.0f, 0.0f, 0.0f));
	la.setColor(0, new Color3f(1.0f, 0.0f, 0.0f));
	la.setCoordinate(1, new Point3f(3*maxCoord/2, 0.0f, 0.0f));
	// la.setCoordinate(1, new Point3f(2.0f, 0.0f, 0.0f));
	la.setColor(1, new Color3f(1.0f, 0.0f, 0.0f));
	// y-axis
	la.setCoordinate(2, new Point3f(0.0f, 0.0f, 0.0f));
	// la.setCoordinate(2, new Point3f(0.0f, 0.0f, 0.0f));
	la.setColor(2, new Color3f(0.0f, 1.0f, 0.0f));
	la.setCoordinate(3, new Point3f(0.0f, 3*maxCoord/2, 0.0f));
	// la.setCoordinate(3, new Point3f(0.0f, 2.0f, 0.0f));
	la.setColor(3, new Color3f(0.0f, 1.0f, 0.0f));
	// z-axis
	la.setCoordinate(4, new Point3f(0.0f, 0.0f, 0.0f));
	// la.setCoordinate(4, new Point3f(0.0f, 0.0f, 0.0f));
	la.setColor(4, new Color3f(0.0f, 0.0f, 1.0f));
	la.setCoordinate(5, new Point3f(0.0f, 0.0f, 3*maxCoord/2));
	// la.setCoordinate(5, new Point3f(0.0f, 0.0f, 2.0f));
	la.setColor(5, new Color3f(0.0f, 0.0f, 1.0f));

	Appearance appr = new Appearance();
	LineAttributes lineAttr = new LineAttributes();
	lineAttr.setLineAntialiasingEnable(true);
	appr.setLineAttributes(lineAttr);	

	axes = new Shape3D(la, appr);
	axesSwitch = new Switch();
	axesSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
	axesSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
	roomTransformGroup.addChild(axesSwitch);
	axesSwitch.addChild(axes);

	drawAxes();
    }

    public void drawAxes()
    {
	if (axesState)
	    axesSwitch.setWhichChild(Switch.CHILD_ALL);
	else
	    axesSwitch.setWhichChild(Switch.CHILD_NONE);
    }

    // TODO: adjustable sizes dynamically?
    // probably want to encode inside some file
    private void initGrid(int inchPerCell)
    {
	float xdiff = maxX - minX;
	float ydiff = maxY - minY;
	roomPlane = new Plane3D(new Point3f(0,0,maxZ),
				new Vector3f(xdiff,0,0),
				new Vector3f(0,ydiff,0),
				(int)(xdiff / inchPerCell),
				(int)(ydiff / inchPerCell),
				new Color3f(1,1,1));

	gridSwitch = new Switch();
	gridSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
	gridSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
	addSwitchToRoom(gridSwitch);
	gridSwitch.addChild(roomPlane);
	drawGrid();
    }

    public void drawGrid()
    {
	if (gridState)
	    gridSwitch.setWhichChild(Switch.CHILD_ALL);
	else
	    gridSwitch.setWhichChild(Switch.CHILD_NONE);
    }

    // TODO: put in labels or something, too, under the same
    // transformgroup nodes
    private void initListenerRadii(HashMap beaconLocations)
    {
	String nextKey;
	String listener1Key;
	String listener2Key;
	Sphere s, s1, s2;
	Appearance appr, appr1, appr2;
	ColoringAttributes colorAttr, colorAttr1, colorAttr2;
	TransparencyAttributes transparencyAttr1, transparencyAttr2;
	TransformGroup tg1, tg2;
	// Shape3D textShape;
	// boolean b = false;

	beaconsSwitch = new Switch();
	beaconsSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
	beaconsSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
	listenerOneRadiiSwitch = new Switch();
	listenerOneRadiiSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
	listenerOneRadiiSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);
	listenerTwoRadiiSwitch = new Switch();
	listenerTwoRadiiSwitch.setCapability(Switch.ALLOW_SWITCH_READ);
	listenerTwoRadiiSwitch.setCapability(Switch.ALLOW_SWITCH_WRITE);

	addSwitchToRoom(beaconsSwitch);
	addSwitchToRoom(listenerOneRadiiSwitch);
	addSwitchToRoom(listenerTwoRadiiSwitch);

	for (Iterator itr = beaconLocations.keySet().iterator();
	     itr.hasNext();)
	    {
		nextKey = (String)(itr.next());
		listener1Key = nextKey + " Listener 1";
		listener2Key = nextKey + " Listener 2";
		
		appr = new Appearance();
		colorAttr = new ColoringAttributes();
		colorAttr.setColor(1.0f, 1.0f, 1.0f);
		appr.setColoringAttributes(colorAttr);
		s = new Sphere(12.0f, appr);
		addPrimitiveToRoomWithSwitch(nextKey, s, beaconsSwitch);
		setPrimitivePositionByName(nextKey,
					   new Vector3f((Vector3d)
							((beaconLocations.get
							  (nextKey)))));

		appr1 = new Appearance();
		colorAttr1 = new ColoringAttributes();
		colorAttr1.setColor(1.0f, 1.0f, 0.0f);
		appr1.setColoringAttributes(colorAttr1);
		
		transparencyAttr1 = new TransparencyAttributes();
		transparencyAttr1.setTransparency(0.7f);
		transparencyAttr1.setTransparencyMode
		    (TransparencyAttributes.BLENDED);
		appr1.setTransparencyAttributes(transparencyAttr1);

		System.out.println("listener1key = " + listener1Key);
		s1 = new Sphere(1.0f, appr1);
		addPrimitiveToRoomWithSwitch
		    (listener1Key, s1, listenerOneRadiiSwitch);
		setPrimitivePositionByName
		    (listener1Key, new Vector3f
		     ((Vector3d)(beaconLocations.get(nextKey))));

		appr2 = new Appearance();
		colorAttr2 = new ColoringAttributes();
		colorAttr2.setColor(0.0f, 1.0f, 1.0f);
		appr2.setColoringAttributes(colorAttr2);
		
		transparencyAttr2 = new TransparencyAttributes();
		transparencyAttr2.setTransparency(0.7f);
		transparencyAttr2.setTransparencyMode
		    (TransparencyAttributes.BLENDED);
		appr2.setTransparencyAttributes(transparencyAttr2);

		s2 = new Sphere(1.0f, appr2);
		addPrimitiveToRoomWithSwitch
		    (listener2Key, s2, listenerTwoRadiiSwitch);
		setPrimitivePositionByName
		    (listener2Key, new Vector3f
		     ((Vector3d)(beaconLocations.get(nextKey))));
		
		/*
		  tg = getPrimitiveTransformGroupByName(nextKey);
		  if (!b)
		  {
		  //worldTransformGroup.addChild(new Text2D(nextKey, new Color3f(1.0f, 1.0f, 1.0f), null, 
		  //				12, 0));
		  // TODO: ugh, need to invert the transform again
		  tg.addChild(new Text2D(nextKey, new Color3f(1.0f, 1.0f, 1.0f), null, 
		  120, 0));
		  b = true;
		  }
		*/

	    }
	drawBeacons();
	drawListenerOneRadii();
	drawListenerTwoRadii();
    }

    public void drawBeacons()
    {
	if (beaconsState)
	    beaconsSwitch.setWhichChild(Switch.CHILD_ALL);
	else
	    beaconsSwitch.setWhichChild(Switch.CHILD_NONE);
    }

    public void drawListenerOneRadii()
    {
	if (listenerOneRadiiState)
	    listenerOneRadiiSwitch.setWhichChild(Switch.CHILD_ALL);
	else
	    listenerOneRadiiSwitch.setWhichChild(Switch.CHILD_NONE);
    }

    public void drawListenerTwoRadii()
    {
	if (listenerTwoRadiiState)
	    listenerTwoRadiiSwitch.setWhichChild(Switch.CHILD_ALL);
	else
	    listenerTwoRadiiSwitch.setWhichChild(Switch.CHILD_NONE);
    }

    // TODO: reread file, or read file upon opening
    // when hitting render, render new file, dump old
    // Ugh, to do that, we need to change the architecture
    // have to have the renderer in the background, and compose
    // and forward the canvas3d, so we can dispose of the old one
    
    private void drawCursor()
    {
	Appearance cursorAppr = new Appearance();
	cursorAppr.setColoringAttributes
	    (new ColoringAttributes(1.0f,1.0f,1.0f,
				    ColoringAttributes.SHADE_FLAT));
	Sphere cursor = new Sphere(12.0f, cursorAppr);
	addPrimitiveToRoom("cursor", cursor);
    }

    private void drawListeners()
    {
	// Listener One
	Appearance listenerOneAppr = new Appearance();
	listenerOneAppr.setColoringAttributes
	    (new ColoringAttributes(1.0f,0.0f,0.0f,
				    ColoringAttributes.SHADE_FLAT));
	Sphere listenerOneSphere = new Sphere(12.0f, listenerOneAppr);
	
	Appearance listenerTwoAppr = new Appearance();
	listenerTwoAppr.setColoringAttributes
	    (new ColoringAttributes(0.0f,0.0f,1.0f,
				    ColoringAttributes.SHADE_FLAT));
	Sphere listenerTwoSphere = new Sphere(12.0f, listenerTwoAppr);

	addPrimitiveToRoom("Listener 1", listenerOneSphere);
	addPrimitiveToRoom("Listener 2", listenerTwoSphere);
    }

    // TODO: reject values outside a certain distance?
    // constraint by lever arm
    private void initLever()
    {
	Appearance appr = new Appearance();
	LineAttributes lineAttr = new LineAttributes();
	lineAttr.setLineAntialiasingEnable(true);
	appr.setLineAttributes(lineAttr);
	
	LineArray la = new LineArray(2,
				     LineArray.COORDINATES |
				     LineArray.COLOR_3);
	la.setCapability(LineArray.ALLOW_COORDINATE_READ);
	la.setCapability(LineArray.ALLOW_COORDINATE_WRITE);

	la.setCoordinate(0, new Point3f(0.0f, 0.0f, 0.0f));
	la.setColor(0, new Color3f(1.0f, 0.0f, 1.0f));
	la.setCoordinate(1, new Point3f(200.0f, 200.0f, 200.0f));
	la.setColor(1, new Color3f(1.0f, 0.0f, 1.0f));

	lever = new Shape3D(la, appr);
	lever.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
	lever.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);

	addShapeToRoom("lever", lever);
    }

    // TODO: want to set laser beam by three parameters
    // front point, direction vector, and distance
    private void initLaserBeam()
    {
	Appearance appr = new Appearance();
	LineAttributes lineAttr = new LineAttributes();
	lineAttr.setLineAntialiasingEnable(true);
	appr.setLineAttributes(lineAttr);

	LineArray la = new LineArray(2,
				     LineArray.COORDINATES |
				     LineArray.COLOR_3);
	la.setCapability(LineArray.ALLOW_COORDINATE_READ);
	la.setCapability(LineArray.ALLOW_COORDINATE_WRITE);

	la.setCoordinate(0, new Point3f(0.0f, 0.0f, 0.0f));
	la.setColor(0, new Color3f(1.0f, 0.141f, 0.0f));
	la.setCoordinate(1, new Point3f(-200.0f, -200.0f, -200.0f));
	la.setColor(1, new Color3f(1.0f, 0.141f, 0.0f));

	laserBeam = new Shape3D(la, appr);
	laserBeam.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
	laserBeam.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);

	addShapeToRoom("laserBeam", laserBeam);
    }

    // TODO: fix this
    private void finishScene()
    {
	BoundingSphere boundingSphere
	    = new BoundingSphere(new Point3d(roomCentroid), 1000);

	universe = new SimpleUniverse(currentCanvas);
	View view = universe.getViewer().getView();
	view.setFrontClipDistance(0.001);

	viewingPlatform = universe.getViewingPlatform();
	viewingPlatform.setNominalViewingTransform();
	TransformGroup viewTransform
	    = viewingPlatform.getViewPlatformTransform();
	Transform3D t = new Transform3D();
	viewTransform.getTransform(t);
	Vector3f temp = new Vector3f(roomCentroid);
	temp.z += 20*maxZ;
 	t.setTranslation(temp);
	t.setScale(900);
	viewTransform.setTransform(t);

	// Keyboard Navigator
	KeyNavigatorBehavior keynav = new KeyNavigatorBehavior(viewTransform);
	keynav.setSchedulingBounds(boundingSphere);

	mouseRotate = new MouseRotate(MouseRotate.INVERT_INPUT);
	mouseRotate.setTransformGroup(viewTransform);
	mouseRotate.setSchedulingBounds(boundingSphere);
	

	mouseTranslate = new MouseTranslate();
	mouseTranslate.setTransformGroup(viewTransform);
	mouseTranslate.setSchedulingBounds(boundingSphere);
	mouseTranslate.setFactor(10,10);

	mouseZoom = new MouseZoom();
	mouseZoom.setTransformGroup(viewTransform);
	mouseZoom.setSchedulingBounds(boundingSphere);
	mouseZoom.setFactor(10);

	scene.addChild(keynav);
	scene.addChild(mouseRotate);
	scene.addChild(mouseTranslate);
	scene.addChild(mouseZoom);
	

	roomTransform = new Transform3D();
	roomTransform.setTranslation(new Vector3f(minX,minY,minZ));
	roomTransformGroup.setTransform(roomTransform);
	scene.addChild(roomTransformGroup);

	scene.compile();
	universe.addBranchGraph(scene);	
    }

    // ================= Communication with GUI ================

    // TODO: render text
    // How many listeners?  Just two for now
    // But should support an arbitrary number of beacons/listeners?

    public synchronized void updatePosition(int listenerNum, Vector3f position)
    {
	switch(listenerNum)
	    {
	    case 1:
		setPrimitivePositionByName("Listener 1", position);
		break;
	    case 2:
		setPrimitivePositionByName("Listener 2", position);
		break;
	    default:
		break;
	    }
    }

    public synchronized void updateBeaconRadii
	(int listenerNum, HashMap distanceEstimates)
    {
	String beaconName;
	Double distance;
	double s;
	TransformGroup tg;
	Transform3D t;
	for(Iterator itr=distanceEstimates.keySet().iterator(); itr.hasNext();)
	    {
		beaconName = (String)(itr.next());
		distance = (Double)(distanceEstimates.get(beaconName));
		if (distance != null)
		    {
			s = distance.doubleValue();
			if (s > 0)
			    {
				tg = getPrimitiveTransformGroupByName
				    (beaconName + " Listener " + listenerNum);
				t = new Transform3D();
				tg.getTransform(t);
				t.setScale(s);
				tg.setTransform(t);
			    }		
		    }
	    }
    }

    // TODO: make them instance variables or something
    public synchronized void updateLever()
    {
	Point3f listener1Position
	    = new Point3f(getPrimitivePositionByName("Listener 1"));
	Point3f listener2Position
	    = new Point3f(getPrimitivePositionByName("Listener 2"));
	LineArray la = (LineArray)(lever.getGeometry());
	la.setCoordinate(0,listener1Position);
	la.setCoordinate(1,listener2Position);
    }

    // TODO: use switches for laser and lever
    public synchronized void updateLaser(String message, String id)
    {
	float distance = 0.0001f;
	try
	    { distance = Float.parseFloat(message) * 39.37f; } // m -> in
	catch (NumberFormatException nfe) {}	    

	System.out.println("distance = " + distance);

	Point3f listener1Position
	    = new Point3f(getPrimitivePositionByName("Listener 1"));
	Point3f listener2Position
	    = new Point3f(getPrimitivePositionByName("Listener 2"));
	// Point3f listener1Position = new Point3f(0,0,0);
	// Point3f listener2Position = new Point3f(-200,-200,-200);
	Vector3f direction
	    = new Vector3f(listener1Position.x - listener2Position.x,
			   listener1Position.y - listener2Position.y,
			   listener1Position.z - listener2Position.z);
	direction.normalize();
	direction.scale(distance);
	Point3f finalPoint = new Point3f(listener1Position);
	finalPoint.add(direction);
	// HACK: laserbeam != null trick
	if (laserBeam != null)
	    {
		LineArray la = (LineArray)(laserBeam.getGeometry());
		la.setCoordinate(0,listener1Position);
		la.setCoordinate(1,finalPoint);
	    }
    }

    public boolean isRendered()
    {
	return rendered;
    }

    public void setAxesState(boolean b)
    {
	axesState = b;
    }

    // TODO: adjustable grid spacing?
    public void setGridState(boolean b)
    {
	gridState = b;
    }
    
    public void setBeaconsState(boolean b)
    {
	beaconsState = b;
    }

    public void setListenerOneRadiiState(boolean b)
    {
	listenerOneRadiiState = b;
    }

    public void setListenerTwoRadiiState(boolean b)
    {
	listenerTwoRadiiState = b;
    }

    public Canvas3D getCanvas3D()
    {
	return currentCanvas;
    }
    
    public void setCanvas3D(Canvas3D canvas)
    {
	currentCanvas = canvas;
    }
}
