import serial
import threading
import time

# Notes: serial.flushInput() flushes the pipe that's coming into the machine
# output is from machine out of serial port

class LaserInterface:

    # TODO: reinitialize??, what to when it doesn't work
    def __init__(self, deviceString='/dev/ttySA0', baudrate=9600,
                 bytesize=serial.SEVENBITS, parity=serial.PARITY_EVEN,
                 stopbits=serial.STOPBITS_ONE, timeout=5,
                 xonxoff=1, rtscts=0):
        self.onState = 0
        self.pulseState = 0
        self.pulseSemaphore = threading.Semaphore(1)
        self.trackingState = 0
        self.trackSemaphore = threading.Semaphore(1)
        self.callbacks = [self.printCallBack]
        self.callbacksSemaphore = threading.Semaphore(1)
        self.callback('Opening: ' + deviceString)
        self.tty = serial.Serial(deviceString, baudrate, bytesize, parity,
                                 stopbits, timeout, xonxoff, rtscts)

    def __del__(self):
        self.tty.flushInput()
        self.tty.flushOutput()
        if self.onState == 1:
            self.off()
        self.tty.close()

    # default callback function - prints to stdout
    def printCallBack(self, value):    
        print value
        
    def on(self):
        if self.onState == 1:
            self.callback('Laser is already on!')
        else:
            try:
                self.tty.write('a')
                self.onState = 1
                self.tty.read(4) # forget the hack using the timer, use blocking i/o
		self.callback('Laser on')
            except serial.serialutil.SerialException:
                self.callback('Unable to switch laser on')

    def off(self):
        if self.onState == 0:
            self.callback('Laser is already off!')
        else:
            try:
                self.tty.write('b\n')
                self.onState = 0
                self.tty.read(3)
		self.callback('Laser off')
            except serial.serialutil.SerialException:
                self.callback('Unable to switch laser off')

    # TODO: rawData[5] is the units
    # 0: mm, 1: 1/100 ft, 6: 1/10 mm, 8: 1/8 in (not so sure about the last one)
    def dist(self):
        if self.onState == 0:
            self.callback('Turn laser on first')
        try:
            self.tty.flushInput()
            self.tty.write('g\n')
            rawData = self.tty.readline() # Error gives '@E255\r\n'
            # units = rawdata[5]
            if rawData[0:5] == '@E255':
                self.callback(str(-1))
            else:
                # TODO: take units, make a nice converter
                try:
                    distValue = int(rawData[7:15])
                    self.callback(str(distValue / 1000.0))
                except ValueError:
                    return
        except serial.serialutil.SerialException:
            self.callback('Communications error')

    def pulseOn(self, seconds):
        self.trackSemaphore.acquire()
        if self.trackingState == 1:
            self.callback('Turn off tracking mode first')
            self.trackSemaphore.release()
            return
        self.trackSemaphore.release()
        self.pulseSemaphore.acquire()
        if self.onState == 0:
            self.callback('Turn laser on first')
        elif self.pulseState == 1:
            self.callback('Pulse mode already on')
        else:
            self.myTimer = threading.Timer(seconds, self.distPulse, [seconds])
            self.pulseState = 1
            self.callback('Pulse mode activated')
            self.myTimer.start()
        self.pulseSemaphore.release()

    def pulseOff(self):
        self.pulseSemaphore.acquire()
        if self.onState == 0:
            self.callback('Turn laser on first')
        elif self.pulseState == 0:
            self.callback('Pulse mode already off')
        else:
            self.pulseState = 0
            self.callback('Pulse mode deactivated')
        self.pulseSemaphore.release()

    def distPulse(self, seconds):
        self.pulseSemaphore.acquire()
        if self.pulseState == 1:
            self.dist()
            del self.myTimer
            self.myTimer = threading.Timer(seconds, self.distPulse, [seconds])
            self.myTimer.start()
        self.pulseSemaphore.release()

    def trackOn(self):
        self.pulseSemaphore.acquire()
        if self.pulseState == 1:
            self.callback('Turn off pulse mode first!')
            self.pulseSemaphore.release()
            return
        self.pulseSemaphore.release()
        if self.onState == 0:
            self.callback('Turn laser on first')
        else:
            if self.trackingState == 1:
                self.callback('Already in tracking mode')
            else:
                self.trackingState = 1
                self.laserTrackingThread = LaserTrackingThread(self)
                self.laserTrackingThread.startTracking()
                self.callback('Tracking on')

    def trackOff(self):
        if self.onState == 0:
            self.callback('Turn laser on first')
        else:
            if self.trackingState == 0:
                self.callback('Laser not in tracking mode')
            else:
                self.trackingState = 0
                # self.tty.flushInput()
                self.callback('Tracking off')

    def register(self, f):
        self.callbacksSemaphore.acquire()
        self.callbacks.append(f)
        self.callbacksSemaphore.release()

    def deregister(self, f):
        self.callbacksSemaphore.acquire()
        self.callbacks.remove(f)
        self.callbacksSemaphore.release()

    def callback(self, value):
        self.callbacksSemaphore.acquire()
        for i in range(0, len(self.callbacks)): # goes from 0 to n-1
            func = self.callbacks[i]
            func(value)
        self.callbacksSemaphore.release()

class LaserTrackingThread(threading.Thread):

    def __init__(self, laserInterface):
        threading.Thread.__init__(self)
        self.laserInterface = laserInterface

        #        def __del__(self):
        # TODO: figure out how to destroy...
        # threading.Thread.__del__(self)
        
    def startTracking(self):
        self.laserInterface.tty.flushInput()
        self.laserInterface.tty.write('h\n')
        self.start()

    def run(self):
        while self.laserInterface.trackingState == 1:
            try:
                self.laserInterface.trackSemaphore.acquire()
                rawData = self.laserInterface.tty.readline()
                if rawData[0:5] == '@E255':
                    self.laserInterface.callback('0.25')
                    self.laserInterface.tty.write('c\n') # stop
                    time.sleep(3) # wait 3 seconds
                    self.laserInterface.tty.readline()
                    self.laserInterface.tty.write('h\n')
                else:
                    # TODO: take units, make a nice converter
                    try:
                        distValue = int(rawData[7:15])
                        self.laserInterface.callback(str(distValue / 1000.0))
                    except ValueError:
                        print 'Value error'
                        # continue
                self.laserInterface.trackSemaphore.release()
            except serial.serialutil.SerialException:
                break
        self.laserInterface.tty.write('c\n')
        self.laserInterface.tty.flushInput()
